#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""image client module."""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, Tuple, overload

from peak.base_client import BaseClient
from peak.constants import ArtifactInfo, ContentType, HttpMethods
from peak.helpers import download_logs_helper, parse_body_for_multipart_request
from peak.session import Session


class Image(BaseClient):
    """Image Client Class."""

    BASE_ENDPOINT_V2 = "image-management/api/v2"

    def __get_artifact_details(self, artifact: Optional[ArtifactInfo] = None) -> Tuple[str | None, list[str] | None]:
        path, ignore_files = None, None

        if artifact:
            path = artifact.get("path")
            ignore_files = artifact.get("ignore_files")

        return path, ignore_files

    @overload
    def list_images(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        status: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_images(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        status: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_images(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        status: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the list of images.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/images/get_api_v2_images>`__

        Args:
            page_size (int | None): The number of images per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            name: (str | None): Image Name or version to search for.
            status (List[str] | None): Filter images on the basis of the status of the latest version.
                Valid values are `not-ready`, `ready`, `in-use`, `deleting`, `delete-failed`.
            scope (List[str] | None): Filter out on the basis of the type of the image - `global` or `custom`.
            last_build_status (List[str] | None): Filter out on the basis of last build status of the latest version.
                Valid values are `building`, `failed`, `success`, `stopped`, `stopping`.
            tags (List[str] | None): Filter out on the basis of the tags attached to the latest version.
            return_iterator (bool): Whether to return an iterator object or list of images for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT_V2}/images/"
        params = {
            "pageSize": page_size,
            "searchTerm": name,
            "status": status,
            "scope": scope,
            "lastBuildStatus": last_build_status,
            "tags": tags,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="images",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    @overload
    def list_image_versions(
        self,
        image_id: int,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        version: Optional[str] = None,
        status: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_image_versions(
        self,
        image_id: int,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        version: Optional[str] = None,
        status: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_image_versions(
        self,
        image_id: int,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        version: Optional[str] = None,
        status: Optional[List[str]] = None,
        last_build_status: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve a list of image versions.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/get_api_v2_images__imageId__versions>`__

        Args:
            image_id (int): The Image id for which we want to get versions.
            page_size (int | None): The number of images per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            version: (str | None): Search by version.
            status (List[str] | None): Filter versions on the basis of their status.
                Valid values are `not-ready`, `ready`, `in-use`, `deleting`, `delete-failed`.
            last_build_status (List[str] | None): Filter out on the basis of last build status of version.
                Valid values are `building`, `failed`, `success`, `stopped`, `stopping`.
            tags (List[str] | None): Filter out on the basis of the tags attached to the version.
            return_iterator (bool): Whether to return an iterator object or list of images for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any] | Dict[str, Any]: An iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform this operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions"
        params = {
            "pageSize": page_size,
            "searchTerm": version,
            "status": status,
            "lastBuildStatus": last_build_status,
            "tags": tags,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="versions",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def create_image(
        self,
        body: Dict[str, Any],
        artifact: Optional[ArtifactInfo] = None,
    ) -> Dict[str, Any]:
        """Create a new image.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/images/post_api_v2_images>`__

        Args:
            body (Dict[str, Any]):  Represents the body to be used to create image. Schema can be found below.
            artifact (Optional[ArtifactInfo]): Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.
                                        Required for image of source `upload`.

        Returns:
            Dict[str, Any]: `buildId`, `imageId`, and `versionId` of the newly created image and the corresponding version.

        SCHEMA:
            Images can be created by three ways:

            .. tabs::

                .. tab:: Upload

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Github

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "branch": "string",
                                    "repository": "string",
                                    "token": "string",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Dockerfile

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfile": "string"
                                },
                        }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            PayloadTooLargeException: The artifact exceeds maximum size.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT_V2}/images/"

        path, ignore_files = self.__get_artifact_details(artifact)
        body = parse_body_for_multipart_request(body)

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body=body,
            path=path,
            ignore_files=ignore_files,
        )

    def create_version(
        self,
        image_id: int,
        body: Dict[str, Any],
        artifact: Optional[ArtifactInfo] = None,
    ) -> Dict[str, Any]:
        """Create a new version.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/post_api_v2_images__imageId__versions>`__

        Args:
            image_id (int): The ID of the image for which we have to create a new version.
            body (Dict[str, Any]):  Represents the body to be used to create image version. Schema can be found below.
            artifact (Optional[ArtifactInfo]): Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.

        Returns:
            Dict[str, Any]: `imageId`, `buildId`, `versionId` and `autodeploymentId` of the newly created version.

        SCHEMA:
            Image versions can be created by three ways:

            .. tabs::

                .. tab:: Upload

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Github

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "branch": "string",
                                    "repository": "string",
                                    "token": "string",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Dockerfile

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfile": "string"
                                },
                        }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            PayloadTooLargeException: The artifact exceeds maximum size.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions"

        path, ignore_files = self.__get_artifact_details(artifact)
        body = parse_body_for_multipart_request(body)

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body=body,
            path=path,
            ignore_files=ignore_files,
        )

    def describe_image(
        self,
        image_id: int,
    ) -> Dict[str, Any]:
        """Retrieve details of a specific image and its latest version.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/images/get_api_v2_images__imageId_>`__

        Args:
            image_id (int): The ID of the image to retrieve.

        Returns:
            Dict[str, Any]: A dictionary containing the details of the image and its latest version.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT_V2}/images/{image_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def describe_version(
        self,
        image_id: int,
        version_id: int,
    ) -> Dict[str, Any]:
        """Retrieve details of a specific version.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/get_api_v2_images__imageId__versions__versionId_>`__

        Args:
            image_id (int): The ID of the image to retrieve.
            version_id (int): The ID of the image to retrieve.

        Returns:
            Dict[str, Any]: a dictionary containing the details of the version.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions/{version_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def update_version(
        self,
        image_id: int,
        version_id: int,
        body: Dict[str, Any],
        artifact: Optional[ArtifactInfo] = None,
    ) -> dict[str, str]:
        """Update an existing image version.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/patch_api_v2_images__imageId__versions__versionId_>`__

        Args:
            image_id (int): The ID of the image to update.
            version_id (int): The ID of the version to update.
            body (Dict[str, Any]):  Represents the body to be used to update image version. Schema can be found below.
            artifact (Optional[ArtifactInfo]):    Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.

        Returns:
            Dict[str, Any]: `imageId`, `buildId`, `versionId` and `autodeploymentId` of the updated version.

        SCHEMA:
            .. tabs::

                .. tab:: Upload

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Github

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "branch": "string",
                                    "repository": "string",
                                    "token": "string",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Dockerfile

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfile": "string"
                                },
                        }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image or version does not exist.
            PayloadTooLargeException: The artifact exceeds maximum size.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.PATCH, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions/{version_id}"

        path, ignore_files = self.__get_artifact_details(artifact)
        body = parse_body_for_multipart_request(body)

        response: Dict[str, Any] = self.session.create_request(
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body=body,
            path=path,
            ignore_files=ignore_files,
        )
        return response

    def create_or_update_image_version(
        self,
        body: Dict[str, Any],
        artifact: Optional[ArtifactInfo] = None,
    ) -> Dict[str, Any]:
        """Create a new image or version if none exists, update version if exists based on image name and version combination.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/images/post_api_v2_images>`__

        Args:
            body (Dict[str, Any]):  Represents the body to be used to create image. Schema can be found below.
            artifact (Optional[ArtifactInfo]): Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.

        Returns:
            Dict[str, Any]: `buildId`, `imageId`, and `versionId` of the newly created image and the corresponding version when no image/version exists.
                            In case when image exists but version doesn't it will return `buildId` and `versionId` of the newly created version along with `imageId` of the existing image.
                            In case when image and version exists, it will return `buildId` for the updated version along with `imageId` and `versionId` of the existing image / version.

        SCHEMA:
            .. tabs::

                .. tab:: Upload

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Github

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "branch": "string",
                                    "repository": "string",
                                    "token": "string",
                                    "dockerfilePath": "string",
                                    "context": "string"
                                },
                        }

                .. tab:: Dockerfile

                    .. code-block:: json

                        {
                            "name": "string (required)",
                            "type": "string (required)",
                            "description": "string",
                            "version": "string",
                            "buildDetails":
                                {
                                    "source": "string",
                                    "buildArguments": [
                                        {
                                            "name": "string",
                                            "value": "string"
                                        }
                                    ],
                                    "secrets": [],
                                    "useCache": "boolean",
                                    "dockerfile": "string"
                                },
                        }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            PayloadTooLargeException: The artifact exceeds maximum size.
            InternalServerErrorException: The server failed to process the request.
        """
        image_name = body.get("name", "")
        version = body.get("version", "")
        response = (
            {} if not len(image_name) else self.list_images(page_size=100, return_iterator=False, name=image_name)
        )
        filtered_images = list(filter(lambda image: image.get("name", "") == image_name, response.get("images", [])))

        if len(filtered_images) > 0:
            image_id = filtered_images[0]["id"]
            versions_response = (
                {}
                if not len(version)
                else self.list_image_versions(
                    image_id=image_id,
                    version=version,
                    page_size=1,
                    return_iterator=False,
                )
            )
            filtered_versions = list(
                filter(lambda ver: ver.get("version", "") == version, versions_response.get("versions", [])),
            )
            if len(filtered_versions) > 0:
                version_id = filtered_versions[0]["id"]
                return self.update_version(image_id=image_id, version_id=version_id, body=body, artifact=artifact)

            return self.create_version(image_id=image_id, body=body, artifact=artifact)

        return self.create_image(body=body, artifact=artifact)

    def delete_image(
        self,
        image_id: int,
    ) -> Dict[None, None]:
        """Delete an Image. All the versions in the image will first go into `deleting` state before actually being deleted.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/images/delete_api_v2_images__imageId_>`__

        Args:
            image_id (int): The ID of the image to delete.

        Returns:
            dict: Empty dictionary object.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT_V2}/images/{image_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def delete_version(
        self,
        image_id: int,
        version_id: int,
    ) -> Dict[None, None]:
        """Delete an Image Version.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/delete_api_v2_images__imageId__versions__versionId_>`__

        Args:
            image_id (int): The ID of the image to delete.
            version_id (int): The ID of the version to delete.

        Returns:
            dict: Empty dictionary object.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform this operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions/{version_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def delete_versions(
        self,
        image_id: int,
        version_ids: List[int],
    ) -> Dict[None, None]:
        """Delete given versions for an image. All the versions will first go into `deleting` state before they and all their associated resources are deleted.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/versions/delete_api_v2_images__imageId__versions>`__

        Args:
            image_id (int): The ID of the image to delete.
            version_ids (List[int]): The IDs of the versions to delete.

        Returns:
            dict: Empty dictionary object.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform this operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/versions"

        params = {
            "versionIds": version_ids,
        }

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params=params,
        )

    @overload
    def list_image_builds(
        self,
        image_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        version_ids: Optional[List[str]] = None,
        build_status: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_image_builds(
        self,
        image_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        version_ids: Optional[List[str]] = None,
        build_status: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_image_builds(
        self,
        image_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        version_ids: Optional[List[str]] = None,
        build_status: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Lists image builds.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/builds/get_api_v2_images__imageId__builds>`__

        Args:
            image_id (int): ID of the image to fetch image build history.
            date_from (str | None): The date after which the image builds should be included (in ISO format).
                It is 90 days from current date by default.
            date_to (str | None): The date till which the image builds should be included (in ISO format).
                It is current date by default.
            page_size (int | None): Number of builds per page.
            page_number (int | None): Page number to retrieve. Only used when return_iterator is False.
            count (int | None): Number of builds required (Ordered by latest to earliest).
                For example, if 5 is provided, it will return last 5 builds.
                It is -1 by default which means it will return all the available builds within the given dates.
            version_ids (List[str] | None): List of version ids to filter builds.
            build_status (List[str] | None): List of build statuses to filter builds.
                Valid values are `building`, `failed`, `success`, `stopped`, `stopping`.


            return_iterator (bool): Whether to return an iterator object or a dictionary of list of image builds. Defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT_V2}/images/{image_id}/builds"

        params: Dict[str, Any] = {
            "dateTo": date_to,
            "dateFrom": date_from,
            "pageSize": page_size,
            "count": -1 if count is None else count,
            "versions": version_ids,
            "buildStatus": build_status,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                params=params,
                response_key="builds",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def get_build_logs(
        self,
        image_id: int,
        build_id: int,
        next_token: Optional[str] = None,
        save: Optional[bool] = False,  # noqa: FBT002
        file_name: Optional[str | None] = None,
    ) -> Dict[str, Any] | None:
        """Retrieve or save the logs of an image build.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/image-management/api-docs/index.htm#/builds/get_api_v2_images__imageId__builds__buildId__logs>`__

        Args:
            image_id (int): The ID of the image.
            build_id (int): The ID of the image build.
            next_token (str | None): The token to retrieve the next set of logs.
            save (bool): Whether to save the logs to a file. Defaults to False.
            file_name (str | None): File name or path where the contents should be saved. Default file name is `image_build_logs_{image_id}_{build_id}.log`.

        Returns:
            Dict[str, Any]: A dictionary containing the logs of the build. If `save` is set to True, the signed URL to download the logs is returned and the logs are saved to a file.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform this operation.
            NotFoundException: The given image or build does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method = HttpMethods.GET
        get_logs_endpoint = f"{self.BASE_ENDPOINT_V2}/images/{image_id}/builds/{build_id}/logs"
        download_logs_endpoint = f"{get_logs_endpoint}/download"

        if save:
            response = self.session.create_request(
                download_logs_endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
            )
            updated_file_name = (
                f"image_build_logs_{image_id}_{build_id}.log" if not file_name or not len(file_name) else file_name
            )
            download_logs_helper(response=response, file_name=updated_file_name)
            return None

        params: Dict[str, Any] = {
            "nextToken": next_token,
        }

        return self.session.create_request(  # type: ignore[no-any-return]
            get_logs_endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params=params,
        )


def get_client(session: Optional[Session] = None) -> Image:
    """Returns an Image Management client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Image: the image client object
    """
    return Image(session)


__all__: List[str] = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Handler for sending requests to the API."""

from __future__ import annotations

import contextlib
import json
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Callable, ClassVar, Dict, Iterator, List, Optional, Tuple, Type, TypeVar, Union

import requests
from requests.adapters import HTTPAdapter
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor, user_agent
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn
from urllib3.util import Retry

import peak.config
from peak.compression import compress, get_files_to_include, print_file_tree
from peak.constants import ContentType, HttpMethods
from peak.exceptions import BaseHttpException, PayloadTooLargeException
from peak.output import Writer
from peak.telemetry import telemetry
from peak.validators import check_file_size

from ._version import __version__

T = TypeVar("T")
Serializable = Dict[str, Union[str, int, float, bool]]
OptionalSerializable = Optional[Serializable]
writer = Writer(ignore_debug_mode=True)


class HandlerRegistryMeta(type):
    """Metaclass for registering all types of Handler classes."""

    REGISTRY: ClassVar[Dict[ContentType, BaseHandler]] = {}

    def __new__(
        cls: "Type[HandlerRegistryMeta]",
        name: str,
        bases: Tuple[Any, ...],
        attrs: Dict[str, Any],
    ) -> HandlerRegistryMeta:
        """This method runs whenever a new class (that uses this class as its metaclass) is defined.

        This method automatically adds the handler classes to its Registry.
        It uses the `CONTENT_TYPE` attribute of the class as key and the class itself as value in registry.
        Ref: https://charlesreid1.github.io/python-patterns-the-registry.html

        Args:
            name (str): Name of the child class
            bases (tuple): Tuple of the child class's inheritance tree
            attrs (dict): Name and value pairs of all the attributes defined in the child class

        Returns:
            HandlerRegistryMeta: the class itself, forward annotated for type checking

        Raises:
            TypeError: if the child class does not have a `CONTENT_TYPE` attribute
        """
        error_invalid_content_type: str = f"Invalid content type for {name} handler"
        new_cls: "HandlerRegistryMeta" = type.__new__(cls, name, bases, attrs)

        content_type: Optional[ContentType] = attrs.get("CONTENT_TYPE")
        try:
            if content_type and ContentType(content_type):
                cls.REGISTRY[content_type] = new_cls()
        except ValueError as err:
            raise TypeError(error_invalid_content_type) from err
        else:
            return new_cls


class _CombinedMeta(HandlerRegistryMeta, ABCMeta):
    """Utility class for combining multiple meta-classes."""


class AuthRetrySession(requests.Session):
    """Session with extra sugar attached."""

    # used in testing, so we can modify the backoff_factor etc. to speed up the tests
    _DEFAULT_RETRY_CONFIG: ClassVar[Dict[str, Any]] = {
        "backoff_factor": 2,
        "total": 5,
        "status_forcelist": [500, 502, 503, 504],
    }

    def _add_retries(self, retry_config: Optional[Dict[str, Any]] = None) -> None:
        if retry_config is None:
            retry_config = self._DEFAULT_RETRY_CONFIG
        adapter = HTTPAdapter(max_retries=Retry(**retry_config))
        self.mount("https://", adapter=adapter)


class HandlerUtils(AuthRetrySession):
    """Utility class for handling requests."""

    @contextlib.contextmanager
    def make_artifact(
        self,
        path: Optional[str],
        body: Dict[str, Any],
        ignore_files: Optional[list[str]],
    ) -> Iterator[MultipartEncoderMonitor]:
        """Create a multipart/form-data encoded file with given body and path as file.

        Args:
            path (Optional[str]): path to the file or folder that will be compressed and used as artifact
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact
            body (Dict[str, Any]): body content to be sent with artifact

        Yields:
            MultipartEncoderMonitor: MultipartEncoderMonitor generator object
        """
        if path:
            with compress(path, ignore_files) as fh:
                check_file_size(fh)
                encoder = MultipartEncoder(
                    {
                        **body,
                        "artifact": (
                            "artifact.zip",
                            fh,
                            "application/zip",
                        ),
                    },
                )
                callback = self._default_callback(encoder)
                monitor = MultipartEncoderMonitor(encoder, callback)
                yield monitor
        else:
            encoder = MultipartEncoder({**body})
            monitor = MultipartEncoderMonitor(encoder)
            yield monitor

    make_artifact.__annotations__["return"] = contextlib.AbstractContextManager

    def _default_callback(self, encoder: MultipartEncoder) -> Callable[[MultipartEncoderMonitor], None]:
        progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
        )
        progress.start()
        bar = progress.add_task("[green]Uploading", total=encoder.len)

        def callback(monitor: MultipartEncoderMonitor) -> None:
            progress.update(bar, completed=monitor.bytes_read, refresh=True)
            progress.refresh()
            if monitor.bytes_read >= monitor.len:
                progress.stop()
                progress.console.clear_live()

        return callback

    @staticmethod
    def parse_args(arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Parse arguments dict and remove the parameters whose values are not provided.

        Args:
            arguments (Dict[str, Any]): dictionary of arguments

        Returns:
            Dict[str, Any]: filtered dictionary where value is not None
        """
        return {k: v for k, v in arguments.items() if v is not None}

    @staticmethod
    def handle_response(response: requests.Response) -> requests.Response:
        """Handles the response from the API.

        Args:
            response (requests.Response): response object from the API

        Returns:
            requests.Response: response of the request.

        # noqa: DAR401
        Raises:
            BaseHttpException: The HTTP exception based on status code
        """
        if 200 <= response.status_code < 300:  # noqa: PLR2004
            return response

        if response.status_code == 413:  # noqa: PLR2004
            raise PayloadTooLargeException(response.json().get("detail", ""))

        raise BaseHttpException.REGISTRY[response.status_code](response.json())


class BaseHandler(ABC, HandlerUtils, metaclass=_CombinedMeta):
    """Abstract base class for all handler type classes."""

    CONTENT_TYPE: ClassVar[ContentType]

    @abstractmethod
    def handle(
        self,
        url: str,
        method: HttpMethods,
        *,
        headers: Dict[str, str],
        params: Optional[Dict[str, Any]],
        body: Optional[Dict[str, Any]],
        path: Optional[str],
        ignore_files: Optional[list[str]],
        request_kwargs: OptionalSerializable,
    ) -> requests.Response:
        """Placeholder handle method."""
        ...


class MultipartFormDataHandler(BaseHandler):
    """Handles requests with multipart/form-data content type."""

    CONTENT_TYPE = ContentType.MULTIPART_FORM_DATA

    def handle(
        self,
        url: str,
        method: HttpMethods,
        *,
        path: Optional[str],
        headers: Dict[str, str],
        body: Optional[Dict[str, Any]],
        params: Optional[Dict[str, Any]] = None,  # noqa: ARG002
        ignore_files: Optional[list[str]] = None,
        request_kwargs: OptionalSerializable = None,
    ) -> requests.Response:
        """Handle multipart/form-data requests.

        Args:
            url (str): url to send the request to
            method (HttpMethods): method to use for the request, e.g. get, post, put, delete
            headers (Dict[str, str]): headers to send with the request
            params (Dict[str, Any]): params to send to the request, not used for multipart/form-data
            body (Dict[str, Any]): body to send to the request
            path (Optional[str]): path to the file or folder that will be compressed and used as artifact
            request_kwargs (OptionalSerializable): extra arguments to be passed when making the request, defaults to None
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact

        Returns:
            requests.Response: response of the request
        """
        with self.make_artifact(path, self.parse_args(body or {}), ignore_files) as monitor:
            headers = {**headers, "Content-Type": monitor.content_type}
            response: Any = getattr(requests, method.value)(url, data=monitor, headers=headers, **request_kwargs)
        return self.handle_response(response)


class ApplicationJsonHandler(BaseHandler):
    """Handles requests with application/json content type."""

    CONTENT_TYPE = ContentType.APPLICATION_JSON

    def handle(
        self,
        url: str,
        method: HttpMethods,
        *,
        headers: Dict[str, str],
        params: Optional[Dict[str, Any]],
        body: Optional[Dict[str, Any]],
        path: Optional[str] = None,  # noqa: ARG002
        ignore_files: Optional[list[str]] = None,  # noqa: ARG002
        request_kwargs: OptionalSerializable = None,
    ) -> requests.Response:
        """Handle application/json requests.

        Args:
            url (str): url to send the request to
            method (HttpMethods): method to use for the request, e.g. get, post, put, delete
            headers (Dict[str, str]): headers to send with the request
            params (Dict[str, Any]): params to send to the request
            body (Dict[str, Any]): body to send to the request
            path (Optional[str]): path to the file or folder that will be compressed and used as artifact, not used in application/json handler
            request_kwargs (OptionalSerializable): extra arguments to be passed when making the request, defaults to None
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact

        Returns:
            requests.Response: response of the request.
        """
        headers = {**headers, "Content-Type": self.CONTENT_TYPE.value}
        response: Any = getattr(requests, method.value)(
            url,
            params=self.parse_args(params or {}),
            json=body,
            headers=headers,
            **request_kwargs,
        )
        return self.handle_response(response)


class Handler:
    """Handler class to handle requests to the API."""

    USER_AGENT: str = user_agent(__package__ or __name__, __version__)

    @staticmethod
    def _pretty_print_request(
        *_: Any,
        content_type: str,
        path: Optional[str] = None,
        ignore_files: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> None:
        writer.write(f"\nSample request: {json.dumps(kwargs, indent=2)}\n")
        if content_type == ContentType.MULTIPART_FORM_DATA.value and path is not None:
            writer.write("\nArtifact file tree:")
            print_file_tree(get_files_to_include(path, ignore_files))

    @telemetry
    def make_request(
        self,
        url: str,
        method: HttpMethods,
        content_type: ContentType,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
        path: Optional[str] = None,
        request_kwargs: Optional[Dict[str, int | bool | str | float]] = None,
        ignore_files: Optional[list[str]] = None,
    ) -> requests.Response:
        """Redirects the request to the appropriate strategy based on the content type.

        Args:
            url (str): url to send the request to
            method (HttpMethods): The HTTP method to use, e.g. get, post, put, delete
            content_type (ContentType): content type of the request
            headers (Dict[str, str]): headers to send with the request
            params (Dict[str, Any]): params to send to the request
            body (Dict[str, Any]): body to send to the request
            path (Optional[str]): path to the file or folder that will be compressed and used as artifact, defaults to None
            request_kwargs(Dict[str, int | bool | str | float] | None): extra arguments to be passed when making the request.
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact

        Returns:
            requests.Response: response json
        """
        headers = {"User-Agent": self.USER_AGENT} if headers is None else {**headers, "User-Agent": self.USER_AGENT}
        params = params or {}
        body = body or {}
        request_kwargs = request_kwargs or {}

        if peak.config.DEBUG_MODE:
            self._pretty_print_request(
                url=url,
                method=method.value.upper(),
                content_type=content_type.value,
                headers=headers,
                params=params,
                body=body,
                path=path,
                ignore_files=ignore_files,
            )
            response = requests.models.Response()
            response._content = b"{}"  # noqa: SLF001
            return response

        return BaseHandler.REGISTRY[content_type].handle(
            url=url,
            method=method,
            headers=headers,
            params=params,
            body=body,
            path=path,
            request_kwargs=request_kwargs,
            ignore_files=ignore_files,
        )


__all__: List[str] = ["Handler", "ApplicationJsonHandler", "MultipartFormDataHandler"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Template module which handles all things related to templates."""

from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import jinja2
import yaml
from jinja2 import Environment
from jinja2.ext import Extension

from peak import exceptions
from peak.helpers import convert_keys_to_snake_case, remove_none_values


def _parse_jinja_template(template_path: Path, params: Dict[str, Any]) -> str:
    """Read, parse and render the Jinja template text."""
    jinja_loader = _CustomJinjaLoader()
    jinja_env = (
        jinja2.Environment(  # TODO: show warning if variable not found in params  # noqa: TD002, TD003, RUF100, FIX002
            loader=jinja_loader,
            autoescape=False,  # noqa: S701
            extensions=[_IncludeWithIndentation],
        )
    )
    jinja_template: jinja2.Template = jinja_env.get_template(str(template_path))
    return jinja_template.render(params, os_env=os.environ)


def load_template(
    file: Union[Path, str],
    params: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
    markdown_data: Optional[Dict[str, str]] = None,
    *,
    convert_to_snake_case: Optional[bool] = False,
) -> Dict[str, Any]:
    """Load a template file through `Jinja` into a dictionary.

    This function performs the following steps:
        * Passes the `YAML` file to be loaded and parsed through **`Jinja`**
        * **`Jinja`** substitutes the variables with their values as they are found in `params`
        * Loads any other files that are referenced using the Jinja `{% include %}` directive, if it is present.
        * Updates the `context` key path within `image` definitions with respect to its relative parent file path.
        * Loads the rendered YAML file into a `dictionary`.

    Args:
        file (Union[Path, str]): Path to the templated `YAML` file to be loaded.
        params (Dict[str, Any] | None, optional): Named parameters to be passed to Jinja. Defaults to `{}`.
        description (str, optional): Description of press entities in markdown format. Defaults to `None`. Note that this parameter will soon be deprecated. Use `markdown_data` instead.
        markdown_data (Dict[str, str] | None, optional): Dictionary containing the markdown data to be inserted into the template.
            The key is a colon-separated string representing the nested key path (e.g., "body:metadata:description"), and the value is the markdown content.
        convert_to_snake_case (Optional[bool], optional): Convert the keys of the dictionary to snake_case. Defaults to `False`.

    Returns:
        Dict[str, Any]: Dictionary containing the rendered YAML file
    """
    params = {} if params is None else params
    file = Path(file)
    template: str = _parse_jinja_template(file, params)
    parsed_data: Dict[str, Any] = yaml.safe_load(template)

    if convert_to_snake_case:
        parsed_data = convert_keys_to_snake_case(parsed_data)

    if description:
        parsed_data["body"]["metadata"]["description"] = description

    if markdown_data:
        for key, markdown_content in markdown_data.items():
            keys = key.split(":")
            target_dict = parsed_data
            for k in keys[:-1]:
                target_dict = target_dict.setdefault(k, {})
            target_dict[keys[-1]] = markdown_content

    return remove_none_values(parsed_data)


class _CustomJinjaLoader(jinja2.BaseLoader):
    """Custom Jinja loader class which handles the include directive.

    Inspired from the jinja2.FileSystemLoader class.
    """

    def __init__(
        self,
        search_path: Optional[List[str]] = None,
        encoding: str = "utf-8",
    ) -> None:
        """Initialize all variables.

        Args:
            search_path (List[str] | None, optional): Path(s) of the directory to search file in.
            encoding (str): Encoding to use when reading files. Defaults to "utf-8".
        """
        if search_path is None:
            search_path = ["."]

        self.search_path: List[str] = [os.fspath(p) for p in search_path]
        self.encoding: str = encoding
        self.root_file_path: Optional[str] = None
        self.build_context_regex: re.Pattern[str] = re.compile(r"^(\s*context\s*:\s+)(.+)$", flags=re.MULTILINE)
        self.seen_files: Set[Path] = set()

    def _update_image_build_context(self, source: str, file_parent_dir: str) -> str:
        """Updates the context key in image definition to the relative path of the file where it is being imported.

        Args:
            source (str): Content of the file
            file_parent_dir (str): Directory where the file is located

        Returns:
            str: Content of the file with updated context value if present.
        """
        if self.root_file_path is None:
            return source

        def substitute(match: re.Match[str]) -> str:
            """Substitute the context key with the relative path of the file where it is being imported."""
            context_path: str = match.group(2)
            context_path = context_path.strip().strip('"').strip("'")
            final_path: str = str(Path(file_parent_dir) / Path(context_path))
            final_relative_path: str = os.path.relpath(final_path, self.root_file_path)
            return f"{match.group(1)}{final_relative_path}"

        return self.build_context_regex.sub(substitute, source)

    def get_source(self, _: jinja2.Environment, template_path: str) -> Tuple[str, str, Callable[[], bool]]:
        """Searches and reads the template file.

        Args:
            _ (jinja2.Environment): Jinja environment variable.
            template_path (str): Path of the template file.

        # noqa: DAR401
        Raises:
            jinja2.TemplateNotFound: The template file is not found on given path.

        Returns:
            Tuple[str, str, Callable[[], bool]]: Tuple containing 3 variables
                1. Content of the template file
                2. Normalized path where the file was found
                3. Function which checks if the file was modified (not useful in this case)
        """
        template_path = template_path.strip()
        for search_path in self.search_path:
            # Use posixpath even on Windows to avoid "drive:" or UNC
            # segments breaking out of the search directory.
            file_path: str = str(Path(search_path) / Path(template_path))

            if Path(file_path).is_file():
                break
        else:
            error_msg: str = f"File does not exist at path: {template_path!r}"
            raise jinja2.TemplateNotFound(error_msg)

        file_path_obj = Path(file_path)
        absolute_file_path: Path = file_path_obj.resolve()

        if absolute_file_path in self.seen_files:
            error_msg = f"Failed to render template, circular include directive found at {absolute_file_path!r}"
            raise exceptions.InvalidTemplateException(error_msg)

        file_parent_dir = str(file_path_obj.parent)
        with Path(file_path).open(encoding=self.encoding) as f:
            contents: str = f.read()
            contents = self._update_image_build_context(contents, file_parent_dir)

        self.seen_files.add(absolute_file_path)
        self.search_path.append(file_parent_dir)
        if self.root_file_path is None:
            self.root_file_path = file_parent_dir

        return contents, os.path.normpath(file_path), lambda: True


class _IncludeWithIndentation(Extension):
    """Override Jinja include directive to preserve indentation.

    Inspired from: https://github.com/stereobutter/jinja2_workarounds
    """

    @staticmethod
    def _include_statement_regex(block_start: str, block_end: str) -> re.Pattern[str]:
        """Get the compiled regex for finding the include directives in template."""
        return re.compile(
            rf"""
            (^.*)
            (?=
                (
                    {re.escape(block_start)}
                    (?P<block_start_modifier> [\+|-]?)
                    (?P<statement>
                        \s*include
                        \s+.*?
                    )
                    (?P<block_end_modifier> [\+|-]?)
                    {re.escape(block_end)}
                )
            )
            .*$
            """,
            flags=re.MULTILINE | re.VERBOSE,
        )

    def preprocess(self, source: str, _: Optional[str], __: Optional[str] = None) -> str:
        """Enclose all include directives.

        For all the regex matches in the text, enclose the include block
        with indent filter blocks and return updated text.
        """
        env: Environment = self.environment

        block_start: str = env.block_start_string
        block_end: str = env.block_end_string
        pattern: re.Pattern[str] = self._include_statement_regex(block_start=block_start, block_end=block_end)
        re.compile("\n")

        def add_indentation_filter(match: re.Match[str]) -> str:
            """Add indent filter to include blocks."""
            content_before_include: str | Any = match.group(1)
            include_statement: str | Any = match.group("statement").replace("indent content", "")

            block_start_modifier: str | Any = match.group("block_start_modifier") or ""
            block_end_modifier: str | Any = match.group("block_end_modifier") or ""

            start_filter: str = (
                f"{block_start + block_start_modifier} filter indent({len(content_before_include)}) {block_end}"
            )
            include_block: str = f"{block_start} {include_statement} {block_end}"
            end_filter: str = f"{block_start} endfilter {block_end_modifier + block_end}"

            return f"{content_before_include}{start_filter}{include_block}{end_filter}"

        return pattern.sub(add_indentation_filter, source)


__all__: List[str] = ["load_template"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Peak apps deployments commands."""

from __future__ import annotations

from typing import Dict, List, Optional

import typer
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.output import Writer
from peak.press.apps import App
from rich.console import Console

app = typer.Typer()
console = Console()

_DEPLOYMENT_ID = typer.Argument(..., help="ID of the App deployment to be used in this operation.")


@app.command(short_help="Create an App deployment.")
def create(
    ctx: typer.Context,
    file: str = args.TEMPLATE_PATH,
    description_file: Optional[str] = args.TEMPLATE_DESCRIPTION_FILE,
    revision_notes_file: Optional[str] = args.REVISION_NOTES_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** an App deployment. This creates all the resources (Example - Workflow, Webapps, etc) described in the App Spec.

    \b
    🧩 ***Input file schema (yaml):***
    ```yaml
    body (map):
        metadata (map):
            name (string | required: false): Name of the deployment. Must be unique within the tenant.
            title (string | required: false): Title of the deployment.
            summary (string | required: false): Summary of the deployment.
            description (string | required: false): Description of the deployment.
            descriptionContentType (string | required: false): Content type of the description. Should be one of "text/plain" or "text/markdown".
            imageUrl (string | required: false): URL of the image to be associated with the app deployment.
            tags (list(map) | required: false):
                - name (string): Name of the tag.
        parameters (map | required: false):
            <blockName> (map): Dictionary of parameters for the block. Here the key is the name of the block.
                <parameterType> (map): Dictionary of parameters. Here the key is the type of the parameter. Accepted values are "build" or "run".
                    <parameterName> (string | boolean | number | list(string)): Value of the parameter. Here the key is the name of the parameter.
        revision (map | required: false):
            notes (string | required: false): Notes for the deployment revision.
        spec (map):
            id (string): ID of the app spec to be deployed.
            release (map | required: false):
                version (string): A valid semantic release version of the app spec.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak apps deployments create /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb"
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if description_file:
        markdown_files["body:metadata:description"] = description_file
    if revision_notes_file:
        markdown_files["body:revision:notes"] = revision_notes_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, app_client.create_deployment)

    with writer.pager():
        response: Dict[str, str] = app_client.create_deployment(**body)
        writer.write(response)


@app.command("list", short_help="List App deployments.")
def list_app_deployments(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    status: Optional[List[str]] = args.STATUS_FILTER_DEPLOYMENTS,
    name: Optional[str] = args.NAME_FILTER,
    sort: Optional[List[str]] = args.SORT_KEYS,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all the App deployments that have been created for the tenant.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments list --sort createdBy:asc,name --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "deploymentCount": 1,
        "deployments": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 10
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.list_deployments(
            status=status,
            name=name,
            sort=sort,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="Describe an App deployment.")
def describe(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** an App deployment with details of its latest revision.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments describe <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "kind": "app",
        "latestRevision": {...},
        "metadata": {...},
        "spec": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments__deploymentId_)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.describe_deployment(deployment_id)
        writer.write(response)


@app.command(short_help="Delete an App deployment.")
def delete(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** an App deployment. This deletes all the resources that were created as a part of the deployment.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments delete <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/delete_v1_apps_deployments__deploymentId_)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.delete_deployment(deployment_id)
        writer.write(response)


@app.command(short_help="Update the App deployment metadata.")
def update_metadata(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    file: str = args.TEMPLATE_PATH,
    description_file: Optional[str] = args.TEMPLATE_DESCRIPTION_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** the App deployment metadata.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
      body (map):
        name (string | required: false): Name of the deployment. Must be unique within the tenant.
        title (string | required: false): Title of the deployment.
        summary (string | required: false): Summary of the deployment.
        description (string | required: false): Description of the deployment.
        descriptionContentType (string | required: false): Content type of the description. Should be one of "text/plain" or "text/markdown".
        imageUrl (string | required: false): URL of the image to be associated with the app deployment.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak apps deployments update-metadata <deployment_id> /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb"
        "kind": "app",
        "latestRevision": {...},
        "metadata": {...},
        "spec": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/patch_v1_apps_deployments__deploymentId_)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if description_file:
        markdown_files["body:description"] = description_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, app_client.update_deployment_metadata)

    with writer.pager():
        response = app_client.update_deployment_metadata(deployment_id, **body)
        writer.write(response)


@app.command(short_help="Redploy an App deployment.")
def redeploy(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Redeploy*** latest revision an App deployment.

    This allows you to redeploy an App deployment that is in a `failed` or `warning` state, provided at least one of its block deployments is also in a `failed` or `warning` state.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments redeploy <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "deploymentId": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "revision": 2
        "revisionId": "7092bd84-c35d-43c1-90ca-7510a1204dcc"
    }
    ```
    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments__deploymentId__redeploy)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.redeploy(deployment_id)
        writer.write(response)


@app.command(short_help="Create an App deployment revision.")
def create_revision(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    file: str = args.TEMPLATE_PATH,
    revision_notes_file: Optional[str] = args.REVISION_NOTES_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** an App deployment revision. This updates the deployment with the specified spec release.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
      body (map):
        parameters (map | required: false):
            <blockName> (map): Dictionary of blocks within the app, for which to provide parameters. Must be a valid block name.
                <parameterType> (map): Dictionary of parameters. Here the key is the type of the parameter. Accepted values are "build" or "run".
                    <parameterName> (string | boolean | number | list(string)): Value of the parameter. Here the key is the name of the parameter.
        release (map):
            version (string): A valid semantic release version of the app spec.
        revision (map | required: false):
            notes (string | required: false): Notes for the deployment revision.

    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak apps deployments create-revision /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "revision": 2
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments__deploymentId__revisions)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if revision_notes_file:
        markdown_files["body:revision:notes"] = revision_notes_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, app_client.create_deployment_revision)

    with writer.pager():
        response: Dict[str, str] = app_client.create_deployment_revision(deployment_id, **body)
        writer.write(response)


@app.command(short_help="Describe an App deployment revision.")
def describe_revision(
    ctx: typer.Context,
    deployment_id: str = args.DEPLOYMENT_ID,
    revision: str = args.REVISION,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** an App deployment revision.

    \b
    ***Note***: Parameters listed in the response are masked if "hideValue" was set to true
        when creating the associated block spec releases.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments describe-revision --deployment-id <deployment_id> --revision <revision>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "createdAt": "2020-01-01T18:00:00.000Z",
        "createdBy": "jane.smith@peak.ai",
        "id": "7092bd84-c35d-43c1-90ca-7510a1204dcc",
        "latestRevision": {...},
        "notes": "This is a new revision"
        "resources": [...],
        "revision": 2,
        "status": "deploying",
        "spec": {...}
        "parameters": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments__deploymentId__revisions__revision_)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.describe_deployment_revision(deployment_id, revision)
        writer.write(response)


@app.command(short_help="List revisions of an App deployment.")
def list_revisions(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    sort: Optional[List[str]] = args.SORT_KEYS,
    status: Optional[List[str]] = args.STATUS_FILTER_DEPLOYMENT_REVISIONS,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all revisions for a given App deployment.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak apps deployments list-revisions "632a4e7c-ab86-4ecb-8f34-99b5da531ceb" --sort createdBy:asc,createdAt --status deployed,deploying --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 10,
        "revisionCount": 1,
        "revisions": [...]
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments__deploymentId__revisions)
    """
    app_client: App = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = app_client.list_deployment_revisions(
            deployment_id,
            sort=sort,
            status=status,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Validators for peak sdk."""
from __future__ import annotations

import os
import tempfile
from typing import Dict, List

from peak.constants import MAX_ARTIFACT_SIZE_MB, MB
from peak.exceptions import FileLimitExceededException, InvalidParameterException


def check_file_size(fh: tempfile.SpooledTemporaryFile[bytes], max_size: float = MAX_ARTIFACT_SIZE_MB) -> None:
    """Check file is smaller than 10MB."""
    file_size: int = _get_file_size(fh)
    if file_size > max_size * MB:
        raise FileLimitExceededException(max_size=max_size)


def validate_action(action: str) -> None:
    """Validate the action provided.

    Args:
        action (str): The action to be validated.

    Raises:
        InvalidParameterException: If the action is not 'read' or 'write'.

    Returns: None
    """
    if action not in ["read", "write"]:
        raise InvalidParameterException(message="The action must be 'read' or 'write'.")


def validate_feature_path(search_result: Dict[str, bool], feature_path: str) -> None:
    """Validate the feature path provided.

    Args:
        search_result (Dict[str, bool]): The search result for the feature path.
        feature_path (str): The feature path to be validated.

    Raises:
        InvalidParameterException: If the feature path is incomplete and more subfeatures need to be specified.

    Returns: None
    """
    if search_result["deeper_levels"]:
        message = f"'{feature_path}' contains more subfeatures. Please specify the complete path."
        raise InvalidParameterException(message=message)


def _get_file_size(fh: tempfile.SpooledTemporaryFile[bytes]) -> int:
    """Get file size in bytes."""
    old_pos: int = fh.tell()
    fh.seek(0, os.SEEK_END)
    file_size_bytes: int = fh.tell()
    fh.seek(old_pos)
    return file_size_bytes


__all__: List[str] = ["check_file_size"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Peak Artifacts service commands."""
from typing import Any, Dict, List, Optional

import typer
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.output import Writer
from peak.resources.artifacts import Artifact

app = typer.Typer(
    help="Manage artifacts used to create resources like images.",
    short_help="Create and manage Artifacts.",
)

_ARTIFACT_ID = typer.Argument(..., help="ID of the artifact to be used in this operation.")

_DOWNLOAD_PATH = typer.Argument(..., help="Path (including filename) where the downloaded file will be stored.")

_DOWNLOAD_VERSION = typer.Option(
    None,
    help="Artifact version to download. If no version is given then latest version is downloaded.",
)

_DELETE_VERSION = typer.Argument(..., help="Artifact version number to delete.")


@app.command("list", short_help="List artifacts.")
def list_artifacts(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all artifacts for the given tenant.

    \b
    📝 ***Example Usage:***<br/>
    ```bash
    peak artifacts list --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "artifactCount": 1,
        "artifacts": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = artifact_client.list_artifacts(page_size=page_size, page_number=page_number, return_iterator=False)
        writer.write(response)


@app.command(short_help="Describe an artifact.")
def describe(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    page_number: Optional[int] = args.PAGE_NUMBER,
    page_size: Optional[int] = args.PAGE_SIZE,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** an artifact with list of its versions.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak artifacts describe "c7575459-e265-4944-a539-1fbb3336799e" --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "description": "description of this artifact",
        "id": "c7575459-e265-4944-a539-1fbb3336799e",
        "name": "my-artifact",
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 1,
        "versionCount": 1,
        "versions": [...]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts__artifactId_)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = artifact_client.describe_artifact(
            artifact_id=artifact_id,
            page_number=page_number,
            page_size=page_size,
        )
        writer.write(response)


@app.command(short_help="Delete an artifact.")
def delete(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** an artifact along with all of its versions.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak artifacts delete "c7575459-e265-4944-a539-1fbb3336799e"
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/delete_api_v1_artifacts__artifactId_)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = artifact_client.delete_artifact(artifact_id=artifact_id)
        writer.write(response, output_type=OutputTypes.json)


@app.command(short_help="Create a new artifact.")
def create(
    ctx: typer.Context,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new artifact. This also creates the first version inside the artifact.

    \b
    🧩 ***Input File Schema(yaml):***<br/>
    ```yaml
      name (str): Name of the artifact.
      description (str): Description of the artifact.
      artifact (map):
        path (str): Path to the artifact.
        ignore_files (list(str) | required: false): Ignore files to use when creating artifact.
    ```

    \b
    📝 ***Example Usage:***
    ```bash
    peak artifacts create '/path/to/body.yaml' --params-file '/path/to/params.yaml'
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "c7575459-e265-4944-a539-1fbb3336799e",
        "version": 1
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/post_api_v1_artifacts)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]
    body = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, artifact_client.create_artifact)

    with writer.pager():
        response: Dict[str, Any] = artifact_client.create_artifact(**body)
        writer.write(response)


@app.command(short_help="Update an artifact's metadata.")
def update_metadata(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** an artifact's metadata.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map): Artifact metadata.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak artifacts update-metadata '/path/to/body.yaml' --params-file '/path/to/params.yaml'
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/patch_api_v1_artifacts__artifactId_)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, artifact_client.update_artifact)

    with writer.pager():
        response: Dict[None, None] = artifact_client.update_artifact(artifact_id, **body)
        writer.write(response)


@app.command(short_help="Download an artifact.")
def download(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    download_path: str = _DOWNLOAD_PATH,
    version: Optional[int] = _DOWNLOAD_VERSION,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
) -> None:
    """***Download*** a specific version for an artifact and save it with the given filename on the local system.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak artifacts download "c7575459-e265-4944-a539-1fbb3336799e" '/path/to/download' --version 1
    ```

    \b
    🆗 ***Response:***
    ```python
    None
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts__artifactId__download)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        artifact_client.download_artifact(artifact_id=artifact_id, download_path=download_path, version=version)


@app.command(short_help="Create a new version of the artifact.")
def create_version(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new version of the artifact.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      artifact (map):
        path (str): Path to the artifact.
        ignore_files (list(str) | required: false): Ignore files to use when creating artifact.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak artifacts create-version "c7575459-e265-4944-a539-1fbb3336799e" '/path/to/body.yaml' --params-file '/path/to/params.yaml'
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "version": 2
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/put_api_v1_artifacts__artifactId_)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, artifact_client.create_artifact_version)

    with writer.pager():
        response: Dict[str, int] = artifact_client.create_artifact_version(artifact_id=artifact_id, **body)
        writer.write(response)


@app.command(short_help="Delete a version of an artifact.")
def delete_version(
    ctx: typer.Context,
    artifact_id: str = _ARTIFACT_ID,
    version: int = _DELETE_VERSION,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** a version of an artifact.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak artifacts delete-version "c7575459-e265-4944-a539-1fbb3336799e" 1
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/delete_api_v1_artifacts__artifactId___version_)
    """
    artifact_client: Artifact = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = artifact_client.delete_artifact_version(artifact_id=artifact_id, version=version)
        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Artifact client module."""
from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ArtifactInfo, ContentType, HttpMethods
from peak.session import Session


class Artifact(BaseClient):
    """Artifact client class."""

    BASE_ENDPOINT = "artifacts/api/v1"

    @overload
    def list_artifacts(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_artifacts(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_artifacts(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve a list of artifacts.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts>`__

        Args:
            page_size (int | None): The number of artifacts per page.
            page_number (int | None): The page number to retrieve. Only used when return_iterator is False.
            return_iterator (bool): Whether to return an iterator object or a list of artifacts, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/artifacts/"
        params = {"pageSize": page_size}

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                params=params,
                response_key="artifacts",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def create_artifact(
        self,
        name: str,
        artifact: ArtifactInfo,
        description: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Create a new artifact.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/post_api_v1_artifacts>`__

        Args:
            name (str): Name of the artifact.
            artifact (ArtifactInfo):    Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.
            description (str | None): A brief description of the artifact.

        Returns:
            Dict[str, Any]: `Id` and `Version` of the created artifact.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            PayloadTooLargeException: The artifact exceeds maximum size.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/artifacts/"
        body: Dict[str, Any] = {"name": name, "description": description}

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body=body,
            path=artifact["path"],
            ignore_files=artifact.get("ignore_files"),
        )

    def describe_artifact(
        self,
        artifact_id: str,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Retrieve details of a specific artifact with list of its versions.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts__artifactId_>`__

        Args:
            artifact_id (str): The ID of the artifact to retrieve.
            page_number (int | None): The page number to retrieve.
            page_size (int | None): The number of versions per page.

        Returns:
            Dict[str, Any]: a dictionary containing the details of the artifact.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}"
        params = {"pageNumber": page_number, "pageSize": page_size}

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params=params,
        )

    def delete_artifact(self, artifact_id: str) -> Dict[None, None]:
        """Delete an artifact with all its versions.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/delete_api_v1_artifacts__artifactId_>`__

        Args:
            artifact_id (str): The ID of the artifact to delete.

        Returns:
            dict: Empty dictionary object.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def create_artifact_version(
        self,
        artifact_id: str,
        artifact: ArtifactInfo,
    ) -> Dict[str, int]:
        """Create a new version of the artifact.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/put_api_v1_artifacts__artifactId_>`__

        Args:
            artifact_id (str): ID of the artifact for which a new version is to be created.
            artifact (ArtifactInfo):    Mapping of artifact attributes that specifies how the artifact will be generated,
                                        it accepts two keys `path`, which is required and `ignore_files` which is optional, and defaults to `.dockerignore`, it is strongly advised that users use `ignore_files` when generating artifacts to avoid copying any extra files in artifact.

        Returns:
            Dict[str, int]: version number.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            PayloadTooLargeException: The artifact exceeds maximum size.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.PUT, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body={},
            path=artifact["path"],
            ignore_files=artifact.get("ignore_files"),
        )

    def update_artifact(self, artifact_id: str, body: Dict[str, Any]) -> Dict[None, None]:
        """Update an artifact's metadata.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/patch_api_v1_artifacts__artifactId_>`__

        Args:
            artifact_id (str): ID of the artifact to be updated.
            body (Dict[str, Any]): dictionary containing new metadata for artifact.

        Returns:
            dict: Empty dict object.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.PATCH, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def delete_artifact_version(self, artifact_id: str, version: int) -> Dict[None, None]:
        """Delete a version of an artifact.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/delete_api_v1_artifacts__artifactId___version_>`__

        Args:
            artifact_id (str): ID of the artifact.
            version (int): Artifact version number to delete.

        Returns:
            dict: Empty dict object.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}/{version}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def download_artifact(self, artifact_id: str, download_path: str, version: Optional[int] = None) -> None:
        """Download a version of the artifact.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/artifacts/api-docs/index.htm#/artifacts/get_api_v1_artifacts__artifactId__download>`__

        Args:
            artifact_id (str): ID of the artifact to download.
            download_path (str): Path (including filename) where the downloaded file will be stored.
            version (int | None): Artifact version to download. If no version is given then latest version is downloaded.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given image does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/artifacts/{artifact_id}"

        endpoint = f"{endpoint}/download" if version is None else f"{endpoint}/{version}/download"

        self.session.create_download_request(
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            download_path=download_path,
        )


def get_client(session: Optional[Session] = None) -> Artifact:
    """Returns an Artifact client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Artifact: the artifact client object
    """
    return Artifact(session)


__all__: List[str] = ["get_client"]
